%% TEXT CATALOGING version 2.2: AUXILIARY PREDICATES
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
/*
AUTHOR: Pascual Julian-Iranzo (Universidad de Castilla-La Mancha, Spain)
OTHER CONTRIBUTORS: Fernando Saenz-Perez (Univ. Complutense de Madrid, Spain)

This software is licensed for research and educational purposes only and it is
distributed with NO WARRANTY OF ANY KIND. You are freely allowed to use, copy
and distribute WN_CONNECT provided that you make no modifications to any of its
files and give credit to their original authors.
*******************************************************************************/


%%%%%%%%%%%
%% Document Identifier Delimiter = [46, 73]    
%% The code list [46, 73] is equivalent to '.I'
%% Usually the identifier is a number

isDocIdentifierDelimiter(Word) :- Word == [46, 73].



%%%%%%%%%%%
%% Text Delimiter = [46, 73]    
%% The code list [46, 87] is equivalent to '.W'

isTextDelimiter(Word) :- Word == [46, 87].


%%%%%%%%%%%
%% read_word(Word,EOF), extracts a Word from the current input stream.
%% EOF is bounded to 'true' if the end of the file is reached. It skips 
%% all symbols which a considered word delimiters.
%% WordDelimiter = [32, 44, 59, 10]
%% Code 32= ' ', 44=',', 59=';', 10 = NewLine

read_word(Word,EOF) :-
   get0(C),
   (C \== -1 -> %% Code -1 = End of File
	(C == 39 -> %% Code 39 = \'  and starts a compound word
                read_compoundWord(Word, EOF)
                ;
		(member(C, [32, 44, 59, 10]) ->
			read_word(Word,EOF)
			;
			read_remainderWord(RemainderWord,EOF), Word = [C|RemainderWord]
		)
	)
        ;
        Word = [],
        EOF=true
    ).


%% read_remainderWord(-Word,-EOF)
%% Reads characters forming the remainder Word until it finds a word delimiter
%%
read_remainderWord(Word,EOF) :- 
   get0(C),
   (C \== -1 -> %% Code -1 = End of File
	(\+(member(C, [32, 44, 59, 10])) ->
		read_remainderWord(RemainderWord,EOF),
                Word = [C|RemainderWord]
		;
		Word = [], EOF=false
	)
        ;
	Word = [], EOF=true
   ).

%% read_compoundWord(-CompoundWord, -EOF)
%% Reads a compound word delimited by \' (code 39).
%% Example: 'oil gas'
%% It reads sequentially all characters until it finds the
%% character \' (code 39) again.
%%
read_compoundWord(CompoundWord, EOF) :-
        get0(C),
        (C \== 39 -> %% Code 39 = \'
            read_compoundWord(RemainderWord, EOF),
            CompoundWord = [C|RemainderWord]
            ;
            CompoundWord = [], EOF=false
        ).




%%%%%%%%%%%
%%% similarToWith(+Mode, +Keyword, +Word1name, -AD)
%%%
%%% Establishes the degree of relation between Keyword and Word1name
%%%
%%% OBSERVATION: Keyword and Word1name can be a simple word or a word term
%%% pattern of the form W[:T[:S]]
%%% Word1name can be a simple word that comes from the file FileName being 
%%% inspected or from a Keyword gloss if we are generating an ontology file.
%%% If we are generating an ontology file based on hyponymy Word1name is a 
%%% full pattern (W:T:S).
%%%
similarToWith(Mode, Keyword:T:S1, Word1name:T:S2, AD) :-
		!,
		similarToWith_aux(Mode, Keyword:T:S1, Word1name:T:S2, AD).

similarToWith(Mode, Keyword:T:S1, Word1name:T, AD) :-
		!,
		similarToWith_aux(Mode, Keyword:T:S1, Word1name:T:_S2, AD).

similarToWith(Mode, Keyword:T:S1, Word1name, AD) :-
		!,
		similarToWith_aux(Mode, Keyword:T:S1, Word1name:T:_S2, AD).

%%%%

similarToWith(Mode, Keyword:T, Word1name:T:S2, AD) :-
		!,
		similarToWith_aux(Mode, Keyword:T:_S1, Word1name:T:S2, AD).

similarToWith(Mode, Keyword:T, Word1name:T, AD) :-
		!,
		similarToWith_aux(Mode, Keyword:T:_S1, Word1name:T:_S2, AD).

similarToWith(Mode, Keyword:T, Word1name, AD) :-
		!,
		similarToWith_aux(Mode, Keyword:T:_S1, Word1name:T:_S2, AD).

%%%%
similarToWith(Mode, Keyword, Word1name:T:S2, AD) :-
		similarToWith_aux(Mode, Keyword:T:_S1, Word1name:T:S2, AD).

similarToWith(Mode, Keyword, Word1name:T, AD) :-
		similarToWith_aux(Mode, Keyword:T:_S1, Word1name:T:_S2, AD).

similarToWith(Mode, Keyword, Word1name, AD) :-
		similarToWith_aux(Mode, Keyword:T:_S1, Word1name:T:_S2, AD).

%%%%%%%%%%%
%%% similarToWith_aux(+Mode, +Keyword, +Word1name, -AD)
%%%
%%% Establishes the relation degree between Keyword and Word1name
%%%
%%% OBSERVATION: Keyword and Word1name must be word terms, i.e., patterns
%%% of the form W:T:S (not simple words)
%%%
similarToWith_aux(Mode, Keyword:T1:S1, Word1name:T2:S2, AD) :-
	(Mode=ont -> (%%% the relation is established by an ontology
				Keyword ~ Word1name = AD);
	 %%% the relation is established by similarity measures
	 Mode=path -> (findall(D, wn_path(Keyword:T1:S1, Word1name:T2:S2, D), LDs),
				max_degree(LDs, AD));
	 Mode=wup -> (findall(D, wn_wup(Keyword:T1:S1, Word1name:T2:S2, D), LDs),
				max_degree(LDs, AD));
	 Mode=lch -> (findall(D, wn_lch(Keyword:T1:S1, Word1name:T2:S2, D), LDs),
				max_degree(LDs, AD));
	 %%% the relation is established by similarity measures based on information content
	 Mode=res -> (findall(D, wn_res(Keyword:T1:S1, Word1name:T2:S2, D), LDs),
				max_degree(LDs, AD));
	 Mode=jcn -> (findall(D, wn_jcn(Keyword:T1:S1, Word1name:T2:S2, D), LDs),
				max_degree(LDs, AD));
	 Mode=lin -> (findall(D, wn_lin(Keyword:T1:S1, Word1name:T2:S2, D), LDs),
				max_degree(LDs, AD));
	 %%% OTHERWISE 
	 write("Undefined mode: "), write(Mode), nl, 
	 fail
	).
 
max_degree([], 0).
max_degree([D|LDs], AD):- max_list([D|LDs], AD).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
in_wordnet(Word:T:S):- !, wn_s(_,_,Word,T,S,_).
in_wordnet(Word:T):- !, wn_s(_,_,Word,T,_,_).
in_wordnet(Word):- wn_s(_,_,Word,_,_,_).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


%%%%%%%%%%%%%%
%%       HIGH ORDER PREDICATES
%%%%%%%%%%%%%%
% applyTo(P, [X1, ..., Xn]) : Builds and executes the atom P(X1, ..., Xn)
%
applyTo(P, Xs) :- Atom =.. [P|Xs], call(Atom).



%%%%%%%%%%%%%%
%%       STATISTICS 
%%%%%%%%%%%%%%
% This is how time and space are measured.

% stat_start(I), Initialization procedure. I is the total number of inferences 
%              (passes via the call and redo ports) since the system was started.
% stat_start(output)
stat_start(I) :- garbage_collect, %%%% Sicstus % nogc, 
	   statistics(runtime,_), statistics(inferences,I).

% stat_end(I), "I" must be the total number of inferences computed by stat_start
% stat_end(input)
stat_end(I) :- statistics(runtime,[_,MSec]), Time is MSec/1000,
          statistics(inferences,I2), Inf is I2 - I, 
          statistics(global_stack,[GBytes,_]), MbG is GBytes/1048576,
          statistics(local_stack,[LBytes,_]),  MbL is LBytes/1048576,
          write('Runtime: '), write(Time), write(' sec   '), 
          write('Num. of inferences: '), write(Inf), write('  '), nl,
          write('MEMORY USAGE: '), 
          write('(Global stack) '), write(MbG), write(' Mb   '), 
          write('(Local stack) '), write(MbL), write(' Mb   '), 
          nl, nl.


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
