%% TEXT CATALOGING version 2.2: genontology module
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
/*
AUTHOR: Pascual Julian-Iranzo (Universidad de Castilla-La Mancha, Spain)
OTHER CONTRIBUTORS: Fernando Saenz-Perez (Univ. Complutense de Madrid, Spain)

This software is licensed for research and educational purposes only and it is
distributed with NO WARRANTY OF ANY KIND. You are freely allowed to use, copy
and distribute WN_CONNECT provided that you make no modifications to any of its
files and give credit to their original authors.
*******************************************************************************/

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%% NOTE.
% TO USE THE PREDICATES IN THIS FILE IT IS NECESSARY TO HAVE ACCESS TO THE
% LIBRARIES:
%%% Michael A. Covington's Efficient Tokenizer (module etu.pl)
%%%
%%% Accessing WordNet you should have automatic access to them (but it is not the case in BPL)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% GENERATION OF ONTOLOGY FILES
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% gen_ontology_file(+ListOfCategories, +File, +Measure)
%%%
gen_ontology_file(ListOfCategories, File, Measure) :-
   gen_ontology_file(ListOfCategories, File, Measure, glosses).

%%% gen_ontology_file(+ListOfCategories, +File, +Measure, +Mode):-
%%% Given a list of categories, ListOfCategories, the name of a file, File,
%%% and the acronym of a measure, Measure (by now [path, wup, lch, res, lin, jcn,
%%% yarm]), it generates a set of proximity equations and stores them into the file File.
%%%
%%% The words of the ListOfCategories must be expressed as patterns "Word:Type:Sense" because 
%%% it is intended that we start from concepts when generating the ontology.
%%%
%%% If Mode=glosses, the ontology is built starting from the glosses of the words in 
%%% ListOfCategories. 
%%%
%%% If Mode= hyponymy, the ontology is built from a subset of the hyponyms 
%%% of the words in ListOfCategories (by default only up to two levels of hyponyms).
%%%
%%% If Mode= hyponymy(Level), the ontology is built from a subset of the hyponyms 
%%% of the words in ListOfCategories. If Level=N, then N levels of hyponyms are generated).
%%%
gen_ontology_file(ListOfCategories, File, Measure, Mode):-
    file_name_extension(_, Extension, File),
    ((Extension = ont) -> File_ont = File;
     (Extension = '')  -> file_name_extension(File, ont, File_ont)
    ),
    (exists_file(File_ont) ->
        write('The file '), write(File), write(' or '), write(File_ont), write(' does exists.'), nl
        ;
        (member(Measure, [path, wup, lch, res, lin, jcn, hso, lesk, vector, yarm]) ->
            open(File_ont, write, OutputStream),
            write(OutputStream,'%% PROXIMITY EQUATIONS: '), 
		  write(OutputStream,' Measure='), write(OutputStream, Measure), 
		  write(OutputStream,'  Mode='), write(OutputStream, Mode), nl(OutputStream),
            gen_equations(ListOfCategories, Mode, Measure, OutputStream),
            close(OutputStream)
            ;
            write(Measure), write(' is not a similarity or relatedness measure.'), nl
        )
    ).



%%% gen_equations(ListOfCategories, Mode, Measure, OutputStream),
%
gen_equations([], _,  _, _).
gen_equations([Category|Categories], Mode, Measure, OutputStream):-
	(in_wordnet(Category) ->
		gen_listOfWords(Mode, Category, ListOfWords),
		gen_equations4category(ListOfWords, Category, Measure, OutputStream)
		;
		write('Warning: the word '), write(Category), write(' is not in WordNet.'), nl
	),
	gen_equations(Categories, Mode, Measure, OutputStream).


%%% gen_listOfWords(+Mode, +Category, -ListOfWords)
%%%
%%% OBSERVATION: For Mode=hyponymy we have set the level of hyponyms to 2.
%%%
gen_listOfWords(glosses, Category, ListOfWords) :- gen_glossList(Category, ListOfWords).
gen_listOfWords(hyponymy, Category, ListOfWords) :- gen_hyponymList(Category, 2, ListOfWords).
gen_listOfWords(hyponymy(Level), Category, ListOfWords) :- gen_hyponymList(Category, Level, ListOfWords).



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% GENERATION OF PROXIMITY EQUATIONS BASED ON HYPONYMS
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%% gen_hyponymList(+Word, +Level, -Hyponym_list)
%%% Generates the list of the Hyponym SynsetIDs of Word. 
%%%
%%% OBSERVATION: To this list is concatenated the proper W_Synset_ID of the Word
%%%
gen_hyponymList(Word, Level , Hyponym_list) :-
	wn_synset_ID_of(Word, W_Synset_ID),
	wn_hyponyms_upto_level(Word, Level, List_SynSetIDs_Hyponyms),
	list_SynSet_2_Hyponym_list([W_Synset_ID|List_SynSetIDs_Hyponyms], Hyponym_list).

%%% list_SynSet_2_Hyponym_list(List_SynSet_IDs, Hyponym_list)
%%% Transform a list of SynSet_IDs into a list of word patterns. 
%%% For each SynSet_ID the word pattern components of the synsent are generated and
%%% they are added to the final list Hyponym_list.
%%%
list_SynSet_2_Hyponym_list(List_SynSet_IDs, Hyponym_list) :-
	list_SynSet_2_Hyponym_list_DL(List_SynSet_IDs, Hyponym_list/[]).

%%% list_SynSet_2_Hyponym_list_DL(+List_SynSet_IDs, -Hyponym_list/Tail)
%%% A difference list implementation of the last predicate.
%%%
list_SynSet_2_Hyponym_list_DL([], Tail/Tail).
list_SynSet_2_Hyponym_list_DL([SynSet_ID|List_SynSet_IDs], Hyponym_list/Tail) :-
	wn_synset_components(SynSet_ID, SynSet_Words),
	phrase(SynSet_Words, Hyponym_list, RestOfHyponym_list), 
	list_SynSet_2_Hyponym_list_DL(List_SynSet_IDs, RestOfHyponym_list/Tail).



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% GENERATION OF PROXIMITY EQUATIONS BASED ON GLOSSES
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%% gen_glossList(+Word, -Gloss_list)
%%%
gen_glossList(Word, Gloss_list) :-
    listify_gloss(Word, GL),
    delete_stop_words(GL, RGL),
    convert_words_to_singular(RGL, SGL),
    delete_repeated_elements(SGL, Gloss_list).


convert_words_to_singular([],[]).
convert_words_to_singular([W|Ws], [SW|SWs]):-
            convert_to_singular(W, SW),
            convert_words_to_singular(Ws, SWs).

%%% listify_gloss(+Word, -Gloss_list)
%%%
listify_gloss(Word, Gloss_list) :-
        wn_gloss_of(Word, Gloss),
        atom_string(Gloss, SGloss),
        open_string(SGloss, Stream),
        tokenizeLine(Stream,TGloss),
        tokensWords(TGloss,Gloss_list).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% delete_stop_words(+Gloss_list, -Pruned_Gloss_list)
%%%
delete_stop_words([],[]).
delete_stop_words([Word|Words], Pruned_Gloss_list) :-
        stop_words(StopWords_list),
        %(lists:member(Word, StopWords_list) ->
        (member(Word, StopWords_list) ->
            delete_stop_words(Words, Pruned_Gloss_list)
            ;
            delete_stop_words(Words, Tail_Pruned_Gloss_list),
            Pruned_Gloss_list = [Word|Tail_Pruned_Gloss_list]
        ).


%%%%%%%%
%%% stop_words(List)
%%% English Stop Words (CSV)
%%% FROM WIKIPEDIA: https://en.wikipedia.org/wiki/Stop_words
%%%
stop_words([a, able, about, across, after, all, almost, also, am, among, an, and, any, are, as, at, be, because, been, but, by, can, cannot, could, dear, did, do, does, either, else, ever, every, for, from, get, got, had, has, have, he, her, hers, him, his, how, however, i, if, in, into, is, it, its, just, least, let, like, likely, may, me, might, most, must, my, neither, no, nor, not, of, off, often, on, only, or, other, our, own, rather, said, say, says, she, should, since, so, some, than, that, the, their, them, then, there, these, they, this, tis, to, too, twas, us, wants, was, we, were, what, when, where, which, while, who, whom, why, will, with, would, yet, you, your]).



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%        LOCAL AXILIARY PREDICATES
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%% gen_equations4category(ListOfWords, Category, Measure, OutputStream)
%%% For each word, "Word", in the list of words, "ListOfWords" it builds the proximity 
%%% equation "Category ~ Word = SimDegree" and it write the proximity equation into the
%%% file named "File".
%%%
%%% The degree "SimDegree" is obtained using the similarity measure "Measure", where "Measure"
%%% can be any of the following acronyms: [path, wup, lch, res, lin, jcn, yarm].
%%%
%%% Words of the list ListOfWords  and categories must be expressed as simple word or 
%%% as patterns: "Word:Type:Sense" or "Category:Type:Sense".
%%%
%%% NOTES: The words "Category" and "Word" must be words of the same part of
%%%        speech (either nouns or verbs) except if the "yarm" measure is used.
%%%
%%% 
gen_equations4category([], _, _, _).
gen_equations4category([Word|ListOfWords], Category, Measure, OutputStream):-
		similarToWith(Measure, Category, Word, SimDegree), 
		((SimDegree > 0) -> 
		%%% then generates the equation
		normalize_degree(Measure, SimDegree, NormalizedDegree),
		write_equation(Category, Word, NormalizedDegree, OutputStream)
		;
		true  %%% else continue %
		),
		gen_equations4category(ListOfWords, Category, Measure, OutputStream).


%%% write_equation(+Category, +Word, + SimDegree, +OutputStream)
%%% If the parameter "Category" is given as a word term (pattern: "Category:Type:Sense") 
%%% it is treated by the first two clauses, but if it is a simple word, it is treated by the
%%% third clause. In either case, it generates the proximity equation: 
%%% "Category~Word=SimDegree".
%%%
%%% NOTE that the generated proximity equation is formed by simple words. This is required 
%%% by BPL, which only relates constant symbols in its proximity equations.
%%%
%%% ACKNOWLEDGMENT: Fernando Saenz-Perez suggested how to solve an output problem with "\'"
%%%                 via the "format/3" predicate.
%%%

write_equation(Category:_:_, Word:_:_, SimDegree, OutputStream) :- !, 
		%% atomic_list_concat([Category, '~', Word, '=', SimDegree, '.'], ProxEqu),
		%% write(OutputStream,ProxEqu), nl(OutputStream).
     	format(OutputStream, '~q~~~q=~w.~n',  [Category, Word, SimDegree]).

write_equation(Category:_:_, Word:_, SimDegree, OutputStream) :- !, 
     	format(OutputStream, '~q~~~q=~w.~n',  [Category, Word, SimDegree]).

write_equation(Category:_:_, Word, SimDegree, OutputStream) :- !, 
     	format(OutputStream, '~q~~~q=~w.~n',  [Category, Word, SimDegree]).

write_equation(Category:_, Word:_:_, SimDegree, OutputStream) :- !, 
     	format(OutputStream, '~q~~~q=~w.~n',  [Category, Word, SimDegree]).

write_equation(Category:_, Word:_, SimDegree, OutputStream) :- !, 
     	format(OutputStream, '~q~~~q=~w.~n',  [Category, Word, SimDegree]).

write_equation(Category:_, Word, SimDegree, OutputStream) :- !, 
     	format(OutputStream, '~q~~~q=~w.~n',  [Category, Word, SimDegree]).

write_equation(Category, Word:_:_, SimDegree, OutputStream) :- !, 
     	format(OutputStream, '~q~~~q=~w.~n',  [Category, Word, SimDegree]).

write_equation(Category, Word:_, SimDegree, OutputStream) :- !, 
     	format(OutputStream, '~q~~~q=~w.~n',  [Category, Word, SimDegree]).

write_equation(Category, Word, SimDegree, OutputStream) :- 
     	format(OutputStream, '~q~~~q=~w.~n',  [Category, Word, SimDegree]).



%%%%%%%%
%%% normalize_degree(+Measure, +Degree, -NormalizedDegree)
%% This predicate normalizes to the range [0, 1] the Degree obtained using 
%% a similarity/relatedness Measure. 
%% The normalization is done with regard to the measure maximum value:
%%
%%        NormalizedDegree = Degree / Measure_max_value
%%
%% Note that, measures path, wup, res, jcn, lin, and yarm have Measure_max_value=1, 
%% therefore, no normalization is necessary.
%%
normalize_degree(path, Degree, Degree):- !.

normalize_degree(wup, Degree, Degree):- !.

normalize_degree(lch, Degree, NormalizedDegree):- !, 
		NormalizedDegree is Degree/3.6888794541139363.

% normalize_degree(res, Degree, Degree):- !. 
% WARNING: Check this value! because Measure_max_value is unknown
%  nl, write('WARNING: Normalization is not checked.'), nl.
% There is not a theoretical result telling us which is the maximum of the Resnik
% Measure. I take 35 as the maximum by experience.
normalize_degree(res, Degree, NormalizedDegree):- !, 
		NormalizedDegree is Degree/35.


% normalize_degree(jcn, Degree, Degree):- !.
% WARNING: Check this value! because Measure_max_value is unknown
%  nl, write('WARNING: Normalization is not checked.'), nl.
% There is not a theoretical result telling us which is the maximum of the Resnik
% Measure. I take 125517.4999985363 as the maximum by experience.
normalize_degree(jcn, Degree, NormalizedDegree):- !, 
		NormalizedDegree is Degree/125517.4999985363.

normalize_degree(lin, Degree, Degree):- !.

normalize_degree(yarm, Degree, Degree):- !.

%%% This mesure is not implemented yet and it is not clear what is its 
%%% Measure_max_value
% normalize_degree(lesk, Degree, NormalizedDegree) :- ????.

%%% This mesure is not implemented yet, but it is known that Measure_max_value =16.
%
normalize_degree(hso, Degree, NormalizedDegree):- 
		NormalizedDegree is Degree/16.



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% delete_repeated_elements(+MultiSet, -Set)
delete_repeated_elements(L1, L2) :- delete_repeated_elements(L1, [], L2).

delete_repeated_elements([], A, A).
delete_repeated_elements([X|R], A, L):-
    %(lists:member(X,A) ->
    (member(X,A) ->
        delete_repeated_elements(R, A, L)
        ;
        delete_repeated_elements(R, [X|A], L)
    ).






